"""
Define SpectraDataFrame class
"""
import pandas as pd
import numpy as np
from scipy import integrate
from typing import Union
from ._dataframe_functions import normalize_to_range, bound_errors


class SpectraDataFrame:
    def __init__(self,
                 data: Union[pd.DataFrame, dict],
                 xname=None):
        """
        Returns Constructs SpectraDataFrame
        :param data: pandas DataFrame or dict
        :param xname: name of column or key containing x-values. By default this is the first column.
        """
        if type(data) is dict:  # Make sure data is a DataFrame
            data = pd.DataFrame(data)
        if xname is None:
            self.xname = data.columns[0]
        else:
            self.xname = xname
        self.df = data
        self._index = -1
        self.x = None
        self.names = None
        self._are_ints_in_names = False
        self._update()

    def __iter__(self):
        return iter(self.names)

    def __getitem__(self, key):
        """Return series of column with given column name or index."""
        if type(key) is int:
            if self._are_ints_in_names:
                try:
                    return self.df[key]
                except KeyError:
                    raise KeyError(f'{key} not found. Selecting by index is\
                     disabled when spectra have integer names')
            else:
                return self.df[self.names[key]]
        else:
            return self.df[key]

    def __setitem__(self, key, value):
        """Adds or changes spectra"""
        if key == self.xname:
            raise KeyError('Key is name of x-axis. Cannot change contents of x-axis this way.')
        else:
            self.df[key] = value
            self._update()

    def __contains__(self, item):
        """Checks if item is a name of any spectra column."""
        return item in self.names

    def _update(self):
        """Update attributes after changing some aspect of self.df."""
        self.x = np.array(self.df[self.xname])

        if self.x[0] > self.x[1]:  # if x axis is decreasing
            self.df = self.df.iloc[::-1]  # reverse order
        self.x = np.array(self.df[self.xname])  # reset self.x
        self.names = self.spectra().columns
        self._are_ints_in_names = False
        for name in self.names:
            if type(name) == int:
                self._are_ints_in_names = True
                break

    def to_csv(self, path, sep=None, header=True):
        """
        Save data as a text file.
        :param path: Path to save data.
        :param sep: Separator to use. (Default ',')
        :param header: Whether to include column names.
        :return: None
        """
        if sep is None:
            sep = ','
        self.df.to_csv(path, sep=sep, header=header, index=False)

    def to_tsv(self, path, sep=None, header=True):
        """
        Save data as a text file.
        :param path: Path to save data.
        :param sep: Separator to use. (Default '\t')
        :param header: Whether to include column names.
        :return: None
        """
        if sep is None:
            sep = '\t'
        self.df.to_csv(path, sep=sep, header=header, index=False)

    def copy(self):
        """Return a copy of SpectraDataFrame object."""
        return SpectraDataFrame(self.df.copy(deep=True))

    def remove(self, names):
        """
        Remove spectra in place.
        :param names: List of column names to remove.
        :return: None
        """
        self.df = self.df.drop(names, axis=1)
        self._update()

    def drop(self, names):
        """
        Returns a SpectraDataFrame with removed columns.
        :param names: List of column names to remove.
        :return: SpectraDataFrame
        """
        return SpectraDataFrame(self.df.drop(names, axis=1))

    def apply_function(self, func, inplace=True):
        """
        Applies a function to all spectra.
        function arguments:
            numpy array with x-values.
            pandas series with spectra values.
        function returns:
            array-like replacement spectra, should be same length as input
        """
        new_df = pd.DataFrame(data={self.xname: self.x})
        for col in self.names:
            new_df[col] = func(self.x, self[col])
        if inplace:
            self.df = new_df
            self._update()
        else:
            return SpectraDataFrame(new_df)

    def crop(self, x1, x2, inplace=True):
        """
        Crops data to range [x1,x2]. (Inclusive range)
        :param x1: x-value for lower bound.
        :param x2: x-value for upper bound.
        :param inplace: Perform operation in-place or return a new instance.
        :return: None or SpectraDataFrame
        """
        bound_errors(self.x, x1, x2)
        if inplace:
            self.df = self.df[self.df[self.xname] <= x2]
            self.df = self.df[self.df[self.xname] >= x1]
            self._update()
        else:
            new_df = self.df[self.df[self.xname] <= x2]
            new_df = new_df[new_df[self.xname] >= x1]
            return SpectraDataFrame(new_df)

    def remove_region(self, x1, x2, inplace=True):
        bound_errors(self.x, x1, x2)
        if inplace:
            self.df = self.df[self.df[self.xname] <= x1 or self.df[self.xname] >= x2]
            self._update()
        else:
            new_df = self.df[self.df[self.xname] <= x1 or self.df[self.xname] >= x2]
            return SpectraDataFrame(new_df)

    def spectra(self):
        """Returns DataFrame with x-axis dropped."""
        return self.df.drop([self.xname], axis=1)

    def mean(self):
        """Returns average spectrum."""
        return np.array(self.spectra().mean(axis=1))

    def std(self):
        """Returns standard deviation at each measurement point."""
        return np.array(self.spectra().std(axis=1))

    def sem(self):
        """Return standard error at each measurement point."""
        return np.array(self.spectra().sem(axis=1))

    def normalize(self, method='default', params=None):
        """
        Normalize the spectra by various methods.
        :param method: str of method to use
        :param params: various params to apply
        :return: None
        """
        if params is None:
            params = {}
        if method in ['default', 'Default']:
            value_range = params['range'] if 'range' in params else (0, 1)
            if value_range[0] > value_range[1]:
                raise ValueError('The first element of value_range should be less than the second element.')
            for col in self.names:
                self.df[col] = normalize_to_range(self.df[col], value_range)
        elif method in ['area', 'Area']:
            area = params['area'] if 'area' in params else 1
            zero = params['zero'] if 'zero' in params else True
            for col in self.names:
                spectra = np.array(self.df[col])
                if zero:
                    spectra = spectra - np.min(spectra)
                self.df[col] = area * spectra * \
                    (1 / integrate.cumtrapz(spectra, self.x, initial=0)[-1])
        elif method in ['mean', 'Mean',
                        'average', 'Average']:
            mean = params['mean'] if 'mean' in params else 1
            zero = params['zero'] if 'zero' in params else True
            for col in self.names:
                spectra = np.array(self.df[col])
                if zero:
                    spectra = spectra - np.min(spectra)
                self.df[col] = spectra * (mean / np.mean(spectra))





import numpy as np

def normalize_to_range(array, R):
    """Returns array normalized to range R."""
    array = array - np.min(array) + R[0]
    return array * (R[1] / np.max(array))


def bound_errors(x, x1, x2):
    if x1 >= x2:
        raise ValueError('x2 must be greater than x1')
    if x1 >= x[-1]:
        raise ValueError('x1 is out of range.')
    if x2 <= x[0]:
        raise ValueError('x2 is out of range')



import numpy as np

def distance(a, b):
    """
    Return distance between two arrays.
    :param a: array 1
    :param b: array 2
    :return: distance between two arrays.
    """
    return np.sqrt(np.sum(np.square(a - b)))


def pick_outliers(sdf, N, region=None, method='max'):
    """
    Return N outliers sorted by distance to avg over given region.
    :param sdf: SpectraDataFrame
    :param N: Number of outliers to return
    :param region: Region to confine measurement.
    :param method: Method to choose outliers by.
    :return:
    """
    new_sdf = sdf.copy()
    if region is not None:
        new_sdf.crop(region[0], region[1])
    avg = new_sdf.mean()
    if method == 'distance':
        values = [distance(avg, np.array(new_sdf[i])) for i in new_sdf]
    elif method == 'max':
        values = [max(np.abs(avg - np.array(new_sdf[i]))) for i in new_sdf]
    else:
        raise ValueError('Unknown method.')
    sortedspectra = [x for _, x in sorted(zip(values, sdf.names))]
    sortedspectra.reverse()
    return sortedspectra[:N]



from .spectradataframe.spectradataframe import SpectraDataFrame
from .plotting.plottingtools import plot
from .spectradataframe.importing import read_csv, read_tsv
from .tools.outlier_removal import pick_outliers
from .tools.interpolation import interpolate
from .tools.baselinefitting import fit_baseline



"""
Define simple tools to plot SpectraDataFrame objects.
"""
import matplotlib.pyplot as plt
from ..spectradataframe.spectradataframe import SpectraDataFrame


def plot(sdf: SpectraDataFrame,
         x_label=None,
         y_label=None,
         title=None,
         dpi=None,
         linewidth=None,
         plot_average=False):
    """
    Simple function to plot the contents of a SpectraDataFrame with Matplotlib.
    :param sdf: SpectraDataFrame
    :param x_label: label for x-axis.
    :param y_label: label for y-axis.
    :param title: title.
    :param dpi: dpi of plot. (Default 150)
    :param linewidth: linewidth for plot. (Default 0.5)
    :param plot_average: Whether to plot the average value over all spectra.
    :return:
    """
    if dpi is None:
        dpi=150
    if linewidth is None:
        linewidth = 0.5
    plt.figure(dpi=dpi)
    ax = plt.subplot()
    x = sdf.x
    spectra = sdf.spectra()
    for col in sdf.names:
        ax.plot(x, spectra[col], lw=linewidth, label='')
    if plot_average:
        ax.plot(x, sdf.mean(), lw=1, color='red', zorder=200, label='Avg')
        ax.legend()
    if title is not None:
        ax.set_title(title)
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    plt.show()


"""
Define functions to import .csv or .tsv files.
"""
import pandas as pd
from .spectradataframe import SpectraDataFrame


def read_csv(path, sep=None, header=None):
    if sep is None:
        sep = ','
    if header is None:
        header = 'infer'
    df = pd.read_csv(path, sep=sep, header=header)
    return SpectraDataFrame(df)


def read_tsv(path, sep=None, header=None):
    if sep is None:
        sep = '\t'
    if header is None:
        header = 'infer'
    df = pd.read_csv(path, sep=sep, header=header)
    return SpectraDataFrame(df)


from distutils.core import setup


setup(
    name='spectraframe',
    version='0.2.0',
    packages=['spectraframe',
              'spectraframe.spectradataframe',
              'spectraframe.plotting',
              'spectraframe.tools'],
    license='MIT License',
    description='Package to work with sets of spectroscopic data in python.',
    author='James Haverstick',
    author_email='jhaverstick99@gmail.com'
)



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from ..spectradataframe.spectradataframe import SpectraDataFrame


def interpolate(sdf: SpectraDataFrame, new_x):
    """
    Returns a SpectraDataFrame with new_x values and spectra linearly interpolated to new_x values.
    :param sdf: SpectraDataFrame to interpolate.
    :param new_x: x values to interpolate to.
    :return: SpectraDataFrame with interpolated y values.
    """
    data = {sdf.xname: new_x}
    for col in sdf.names:
        func = interp1d(sdf.x, sdf[col])
        data[col] = np.fromiter(map(func, new_x), float)
    return SpectraDataFrame(pd.DataFrame(data))



import pandas as pd
from scipy.optimize import curve_fit
from ..spectradataframe.spectradataframe import SpectraDataFrame
import numpy as np


def fit_baseline(sdf, func, regions=None, p0=None, bounds=(-np.inf, np.inf)):
    baseline_data = {sdf.xname: sdf.x}
    corrected_data = {sdf.xname: sdf.x}
    params = {}
    sdfcopy = sdf.copy()
    if regions is not None:
        sdfcopy.crop(regions[0][0], regions[-1][1])
        if len(regions) == 1:
            pass
        else:
            for i in range(len(regions) - 1):
                sdfcopy.remove_region(regions[i][1], regions[i+1][0])
    for name in sdf.names:
        popt, pcov = curve_fit(func,
                               sdfcopy.x,
                               sdfcopy[name],
                               bounds=bounds,
                               p0=p0)
        baseline_data[name] = np.array([func(i, *popt) for i in sdf.x])
        corrected_data[name] = sdf[name] - baseline_data[name]
        params[name] = popt
    return SpectraDataFrame(pd.DataFrame(corrected_data)), \
           SpectraDataFrame(pd.DataFrame(baseline_data)), params
